Widget type matching
---------------------

Currently toolkits match the widget type at run-time via `Widget::class()`.
It would be nice to enable compile-time matching.

Proposal: toolkits add their own `Widget` class, and implement it for each
widget type directly.

Problem: toolkits must support all widgets implementing `kas::Widget`,
which makes the above impossible, since only a blanket implementation can be
guaranteed to implement all of the above.

Solution via specialisation: use a blanket implementation *and* more specific
implementations. The blanket impl may panic in some cases ("unsupported widget").

Alternative (run-time only): use a `Downcast<T>` trait like gtk-rs.
We will likely need this anyway; in other words specialisation provides only
a performance boost (unless we never support dynamic objects).

### Widget type specific interfaces

At some point it seems likely we'll want custom functionality depending on the
widget type (currently `label` is universal to avoid this). Most widget toolkits
make use of object-oriented class extension to enable this. Since Rust doesn't
have classes, we will probably need a large number of traits, often associated
with only a single implementation (or at least one most common implementation).
Because of this we will likely want some common prefix/suffix, like GTK-rs uses
the `Ext` suffix.

Up-cast: as soon as Rust supports trait-object upcast we will get this for free;
in the mean-time we need `upcast` or `as_parent_type` methods where trait
object upcast is needed.

Down-cast: for static types this can be achived via a generic function. For
trait objects, the equivalent would require a separate method for each desired
downcast target. gtk-rs has an interesting solution, but does not apply directly
to Rust's fat pointers. In theory downcasting a fat pointer should be as simple
as transmutation and replacing the vtable pointer with the appropriate one, but
there's no API or stable representation for fat pointers; this would be fragile.
Note that we can use `type_id` to identify types.
Alternatively we can just use `Any` or `mopa`.


Dialogs
--------

Direct support for dialogs is critical: newer toolkits like GTK3+ implement some
dialogs like file choosers by sending a DBus message to the window manager /
compositor, since sandboxing is supposed to prevent apps from reading arbitrary
files without user intervention.

We should have a `Dialog` trait or similar; we then need either `Dialog::class()`
or specialisation as above.

It should be possible to "run" a dialog: `toolkit.run(my_dialog)`. Should it
also be possible to add a dialog as a window? Should `trait Dialog: Window`?


Derived widget construction
---------------------

This is largely done via macros to reduce boilerplate; in particular, via
`make_widget!`. This macro:

-   supports multiple layouts: horizontal, vertical, grid
-   allows specification of child widgets, each requiring:
    
    -   position (for grid layout only)
    -   name
    -   type: specified or generic, optionally with message result type
    -   initialiser
    -   handler for child-widget messages
-   allows data fields with name, type and initialiser
-   has a response type

This form of message handling is very messy!

Types of field:

-   data: concrete type + initialiser
-   data: auto type + initialiser?  — but without type bounds this is useless
-   widget: concrete type + initialiser + trivial handler
-   widget: concrete type + initialiser + custom handler
-   widget: auto type + concrete response + initialiser + trivial handler
-   widget: auto type + concrete response + initialiser + custom handler
-   widget: auto type + auto response + initialiser + trivial handler ?


Event handling
--------------

Widgets should:

-   be allowed internal state which is directly accessible
-   be able to define custom handlers for input events
-   be able to send messages directly to the parent (preferably after the
    handler exits to avoid re-entrancy complications)
-   be able to send messages to a central data model (again, preferably after
    the handler exits)
-   be able to access child widgets directly

### Type safety of messages

Messages to the parent: there are only two ways to achieve this; restrictions
between child widget and parent message type, or user specification of a global
message type. The latter is incompatible with API boundaries.

It might be possible to achieve the former by allowing widgets to specify a
message type which child widgets can send, however then type-safe logic
connecting the child's message to the parent's receiver must be customised on
the parent widget's type; this pretty-well requires that event handling is
recursive (each widget passing events to the appropriate child).

Messages to a central data model: this requires user specification of the
message type for this data model; this is also incompatible with API boundaries
but this may not be a problem since presumably only the end application will
define handlers interacting with the data model.

### Non-type-safe messages

If message types are not specific to widgets, an external message handler may
invoke widget event handlers, then if necessary find the parent and call its
receive-from-child handler. However, this still requires a mechanism to find
the parent (e.g. a parent-pointer stored for each child).

### Recursive handling

Model:

1.  input event is sent to widget
2.  pre-passing hook may directly handle or filter
3.  internal logic is used to find the appropriate child, if any
4.  (optional) pre-passing hooks may interact with the event and abort passing
5.  event is passed to child and result obtained
6.  post-passing hook handles result and yields result for this widget

This requires re-implementing most event handling logic instead of relying on
an external toolkit to find the appropriate sub-widget. This may even be
difficult or impossible to implement for some toolkits.

It then leaves the question of whether to force users to write the entire event
handler (if not using the default) or which hooks to allow the user to specify.

---

Enumerate all widgets; allows easy logic to locate a sub-widget from a number.

Indirect event handling: identify widget number, then handle from root.
